#####
###  MailBoxer - a mailinglist/newsletter/mailarchive-framework for ZOPE
##   Copyright (C) 2002	Maik Jablonski (maik.jablonski@uni-bielefeld.de)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
##   along with this program; if not, write to the Free Software
###  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#####

from DateTime import DateTime
from rfc822 import parseaddr
from re import search,match,sub
from Acquisition import Implicit
from Globals import Persistent,InitializeClass,DTMLFile
from AccessControl import ClassSecurityInfo
from OFS.Folder import Folder
from zLOG import LOG, INFO
from smtplib import SMTP
from MailBoxerTemplates import MailBoxerTemplates,MailBoxerPageTemplates

manage_addMailBoxerForm = DTMLFile('dtml/manage_addMailBoxerForm', globals())

def manage_addMailBoxer (self, id, title='', REQUEST=None):
	""" Add a new MailBoxer to current ObjectManager:
	    - create MailBoxer-Folder
	    - create a MailHost
            - copy some DTML-Methods to MailBoxer-Folder
	"""
	
	mb = MailBoxer(id,title,REQUEST)
	self.Destination()._setObject(id,mb)
	mb=self._getOb(id)

	# Add a MailHost to MailBoxer

	mb.manage_addProduct['MailHost'].manage_addMailHost('MailHost',smtp_host=REQUEST.get('smtphost','127.0.0.1'))

	# Copy Template-Methods to mailBoxer
	
	for key in MailBoxerTemplates.keys():
		mb.addDTMLMethod(key,file=MailBoxerTemplates[key])

	for key in MailBoxerPageTemplates.keys():
		mb.manage_addProduct['PageTemplates'].manage_addPageTemplate(key,title='',text=MailBoxerPageTemplates[key])
		
	# Redirect if requested TTW
	
	if REQUEST is not None:
		return self.manage_main(self,REQUEST,update_menu=1)
		

class MailBoxer(Implicit,Persistent,Folder):
	""" MailBoxer-Class - 
	    a simple Folder with some properties,
	    a log,
	    and some methods: securityCheck, manage_inboxer & manage_mailboxer
	"""

	security=ClassSecurityInfo()
	
	meta_type='Mail Boxer'

	_properties=(
			{'id':'title',      'type':'string', 'mode':'w'},
			{'id':'maillist',   'type':'lines',  'mode':'w'},
			{'id':'mailto',     'type':'string', 'mode':'w'},
			{'id':'moderator',  'type':'string', 'mode':'w'},
			{'id':'moderated',  'type':'boolean','mode':'w'},
			{'id':'archived',   'type':'boolean','mode':'w'},
			{'id':'subscribe',  'type':'string', 'mode':'w'},
			{'id':'unsubscribe','type':'string', 'mode':'w'},
			{'id':'mtahosts',   'type':'tokens', 'mode':'w'},
			{'id':'spamlist',   'type':'lines',  'mode':'w'},
			{'id':'history',    'type':'int',    'mode':'w'},
			{'id':'xmailer',    'type':'string', 'mode':'w'}
		    )

	# Add manage_MailBoxerLog
	
	security.declareProtected('View management screens','manage_MailBoxerLog')
	
	manage_MailBoxerLog = DTMLFile('dtml/manage_MailBoxerLog', globals())
	manage_options=Folder.manage_options+({'label':'Log','action':'manage_MailBoxerLog'},)

	# Here's the constructor for the MailBoxer
	
	def __init__(self,id,title='',REQUEST=None):
		self.id           = id
		self.title        = title
		self.maillist     = []
		self.spamlist     = []
		self.log          = []
		self.mailto       = REQUEST.get('mailto','')
		self.moderator    = REQUEST.get('moderator','')
		self.moderated    = REQUEST.get('moderated',0)
		self.archived     = REQUEST.get('archived',0)
		self.mtahosts	  = REQUEST.get('mtahosts',[])
		self.subscribe    = 'subscribe|.*: subscribe'
		self.unsubscribe  = 'unsubscribe'
		self.xmailer      = 'MailBoxer'
		self.history      = 100


	security.declarePublic('securityCheck')
	
	def securityCheck(self,REQUEST):
		""" Checks if request was called from allowed MTA's.
		    This ensures, that no-one can send us mail
		    directly TTW. If you leave mtahosts blank,
		    all hosts will be allowed => means no securityCheck at all!
		"""

		if self.mtahosts and REQUEST['REMOTE_ADDR'] not in self.mtahosts:
		   message = 'Host %s is not allowed' % (REQUEST['REMOTE_ADDR'])
		   LOG('MailBoxer', INFO, message)
		   raise(message)


	security.declarePublic('manage_inboxer')

	def manage_inboxer(self,REQUEST):
		""" inboxer stores mails in
		    yyyy/yyyy-mm/yyyy-mm-dd/-folders

		    Needs a REQUEST with:
		     - subject, from, body
		     - optional: attachments
		       (list of records with: filename,filetype,filecontent)

		    Returns the new created MailFolder-Object
		"""

		self.securityCheck(REQUEST)

		time=DateTime()

		# define the properties for our mailobject

		mail = {
		  'mailSubject' : (REQUEST.get('subject','No subject')  , 'string'),
		  'mailFrom'    : (REQUEST.get('from','No sender')      , 'string'),
		  'mailBody'    : (REQUEST.get('body','No body')        , 'text'),
		  		  		  
		# some mailers (e.g. pine) send us strange datestamps,
		# so it's easier to use our own...

		  'mailDate'    : (time, 'date')
		}

		# mailTitle will be used for the title of the mailobject
		# format: subject / from

		mailTitle=mail['mailSubject'][0] + ' / ' + mail['mailFrom'][0]

		# now let's check/create the date-path (yyyy/yyyy-mm/yyyy-mm-dd) 
		# define the folder-id's along the date

		year  = str(time.year())              # yyyy
		month = year  + '-' + str(time.mm())  # yyyy-mm
		day   = month + '-' + str(time.dd())  # yyyy-mm-dd

		# do we have a year folder already?

		if hasattr(self,year)==0:
		   self.manage_addFolder(year,title=year)

		# get the yearFolder as object

		yearFolder=getattr(self,year)

		# do we have a month folder already?

		if hasattr(yearFolder,month)==0:
		   yearFolder.manage_addFolder(month,title=month)

		# get the monthFolder as object

		monthFolder=getattr(yearFolder,month)

		# do we have the day folder already?

		if hasattr(monthFolder,day)==0:
		   monthFolder.manage_addFolder(day,title=day)

		# get the dayFolder as object

		dayFolder=getattr(monthFolder,day)

		# let's create the mailObject
		
		mailFolder = dayFolder
		
		# maybe it's a reply =>
		# then get the "replied-mail"-folder for storing the reply
		
		for folder in dayFolder.objectValues(['Folder']):
                     if mail['mailSubject'][0].find(folder.mailSubject)<>-1:
                          mailFolder=folder
			  break
			  
		# search a unique id for the mailobject

		# get the number of objects in folder
		# => maybe already a free id?
		
		id = len(mailFolder.objectIds())    

		# to be sure: test and search for a free id...
		
		while hasattr(mailFolder,str(id)):  
		     id=id+1

		mailId=str(id)

		mailFolder.manage_addFolder(mailId,title=mailTitle)
		mailObject=getattr(mailFolder,mailId)

		# and now add some properties to our new mailobject 

		for mailProperty in mail.keys():
		    mailObject.manage_addProperty(
		    				  mailProperty,
                                		  mail[mailProperty][0],
                                		  mail[mailProperty][1]
						 )

		# attachments? Store them as file-objects with leading numbers
		# => a easy way to avoid overriding standard_*-methods
		
		if REQUEST.has_key('attachments'):
			counter=0
			for file in REQUEST['attachments']:
				mailObject.manage_addFile(
							  id=str(counter)+file.filename,
							  title=file.filename,
							  file=file.filecontent,
							  content_type=file.filetype
							 )
				counter=counter+1


	security.declarePublic('manage_logboxer')
		
	def manage_logboxer(self,REQUEST,status='UNPROCESSED'):
		""" Add a MailBoxer-log entry """

		self.securityCheck(REQUEST)
		
		self.log = (
			    [{
				'mailDate'    : str(DateTime().ISO()),
				'mailStatus'  : status,
				'mailSender'  : REQUEST.get('from','No sender'), 
				'mailSubject' : REQUEST.get('subject','No subject')
			    }] + self.log[:self.history-1]
			   )
	
	
	security.declarePublic('manage_mailboxer')

	def manage_mailboxer(self,REQUEST):
		""" MailBoxer manages mailinglists """

		self.securityCheck(REQUEST)

		# mail already processed? avoid endless loops

		if REQUEST.get('x-mailer') == self.xmailer:
			return

		# get raw-address, subject and body

		(RealName, RawAddress)   = parseaddr(REQUEST.get('from','No sender'))
		Subject                  = REQUEST.get('subject','No subject')
		Body                     = REQUEST.get('body','No body')

		# lower-case address for comparing to maillist

		LowAddress = RawAddress.lower()

		# lower-case the maillist for comparison
		
		LowList=[]
		for addr in self.maillist:
			LowList.append(addr.lower())

		# sender is a spammer? log this and return.

		for regexp in self.spamlist:
			if regexp<>'' and search(regexp,LowAddress):
				self.manage_logboxer(REQUEST,status='SPAM-MATCH: %s' % (regexp))
				return			
		
		# subscription? only subscribe if subscription is enabled.

		if self.subscribe<>'' and match(self.subscribe,Subject):
			if LowAddress not in LowList:
				self.maillist.append(RawAddress)
				self.maillist.sort()
				self.maillist=self.maillist
				self.mail_subscribe(self,REQUEST,mailfrom=RawAddress)
				self.manage_logboxer(REQUEST,status='SUBSCRIBE')
			return

		# unsubscription? only unsubscribe if unsubscription is enabled...

		if self.unsubscribe<>'' and match(self.unsubscribe,Subject):
			if LowAddress in LowList:
				index=LowList.index(LowAddress)
				self.maillist=self.maillist[:index]+self.maillist[index+1:]
				self.mail_unsubscribe(self,REQUEST,mailfrom=RawAddress)
				self.manage_logboxer(REQUEST,status='UNSUBSCRIBE')
			return

		# traffic for our list: relay if it's a mail of the moderator
		# or check if the sender is subscribed to the list.
		
		if ((LowAddress==self.moderator.lower()) or
		    (LowAddress in LowList and not self.moderated)):

			# put mail into the archive

			if self.archived:
				self.manage_inboxer(REQUEST)

			messageText=REQUEST.get('rawData')

			# insert additional message-headers/footers
			
			if hasattr(self,'mail_header'):
				mailHeader=self.mail_header(self,REQUEST).strip()
			else:
				mailHeader="X-Mailer: %s\nPrecedence: Bulk" % self.xmailer
			
			if hasattr(self,'mail_footer'):
				mailFooter='\n'+self.mail_footer(self,REQUEST).strip()
			else:
				mailFooter=''
				
			if messageText.find('\r\n\r\n')<>-1:
				messageText=messageText.replace('\r\n\r\n','\r\n%s\r\n\r\n' % mailHeader,1)
			else:				
				messageText=messageText.replace('\n\n','\n%s\n\n' % mailHeader,1)

			# modify subject: insert list title
			
			if Subject.find("[%s]" % self.title)==-1:
				messageText=sub('[Ss]ubject:','Subject: [%s]' % (self.title), messageText,1)

			# remove possible blank lines in maillist...
			
			while '' in self.maillist:
				self.maillist.remove('')
				
			self.maillist=self.maillist

			# bulk mail to all members...
			# it's really fast because we talk directly to SMTP
			
			smtpserver = SMTP(self.MailHost.smtp_host, int(self.MailHost.smtp_port))					
			errorList=smtpserver.sendmail(self.moderator, self.maillist, messageText+mailFooter)
			smtpserver.quit()
						
			self.manage_logboxer(REQUEST,status="BULK TO %d MEMBERS %s" % (len(self.maillist)-len(errorList),str(errorList)))
			return

		# if all previous tests fail,
		# it must be a mail to a moderated list
		# and/or an unknown sender. be kind and give a feedback.
		
		self.mail_reply(self,REQUEST,mailfrom=RawAddress)

		if self.moderated:
			self.manage_logboxer(REQUEST,status="MESSAGE TO MODERATOR")
		else:
			self.manage_logboxer(REQUEST,status="SENDER UNKNOWN")
		return
								

	def nextObject(self,context):
		""" Returns next Folder in context or None """
		parrentObjects = context.aq_parent.objectValues('Folder')

		if context in parrentObjects:
			parrentPosition = parrentObjects.index(context)
			if parrentPosition+1<len(parrentObjects):
				return parrentObjects[parrentPosition+1]
   
		return None

	def previousObject(self,context):
		""" Return previous Folder in context or None """
		parrentObjects = context.aq_parent.objectValues('Folder')

		if context in parrentObjects:
			parrentPosition = parrentObjects.index(context)
			if parrentPosition>0:
				return parrentObjects[parrentPosition-1]
		return None

# And here we go to the moon... Houston, we have a problem...

InitializeClass(MailBoxer)
